function  [U_opt PD_opt sigmatopt alphah_opt x_ce optscore] = LHSS(x_de, x_nu, m, nBasis)
%
% Least-squares Hetero-distributional Subspace Search (LHSS)
%
% Estimating ratio of probability densities
%   \frac{ p_{nu}(x) }{ p_{de}(x) }
% from samples
%    { xde_i | xde_i\in R^{d} }_{i=1}^{n_{de}}
% drawn independently from p_{de}(x) and samples
%    { xnu_i | xnu_i\in R^{d} }_{i=1}^{n_{nu}}
% drawn independently from p_{nu}(x).
%
% Usage:
%       [U_opt PD_opt sigmatopt alphah_opt x_ce optscore] = LHSS(x_de, x_nu, m, nBasis)
%
% Input:
%    x_de:          d by n_de sample matrix corresponding to `denominator' (iid from density p_de)
%    x_nu:          d by n_nu sample matrix corresponding to `numerator'   (iid from density p_nu)
%    nBasis:        Number of basis functions in uLSIF
%    m:             Number of dimensionality
%
%
% Output:
%    nBasis:        Number of basis functions in uLSIF
%    U_opt:         Transformation matrix to Hetero-distributional subspace
%    PD_opt:        Pearson Divergence score
%    sigmatopt:     Gaussian width in uLSIF
%    alphah_opt:    uLSIF parameters
%    x_ce:          Basis vectors
%    optscore:         CV score
%
% (c) Makoto Yamada, Department of Compter Science, Tokyo Institute of Technology, Japan.
%     yamada@sg.cs.titech.ac.jp,
%     http://sugiyama-www.cs.titech.ac.jp/~sugi/software/D3LHSS/

rand('state', 0);
randn('state', 0);

if nargin<2
    error('number of input arguments is not enough!!!')
end

[dim,n_de]=size(x_de);
[dim_nu,n_nu]=size(x_nu);
if dim~=dim_nu
    error('dimension of two samples are diferent!!!')
end


U =  ones(dim,m);
[Q R] = qr(U);
Q = Q*sign(R(1,1));
U = Q(:,1:m);
N_nu = size(x_nu,2);
N_de = size(x_de,2);

nind = randperm(N_nu);
x_ce = x_nu(:,nind(1:nBasis));

temp = U'*x_nu;
med = compmedDist(temp');

[alphah,sigmat,PD,wh_x_de,score]=uLSIF(U'*x_de,U'*x_nu,U'*x_ce,med*[0.75 1.0 1.25]);

sigmatopt = sigmat;
alphah_opt = alphah;

PD_opt = PD;
U_opt = U;
sigma = sigmat^2;

decPDcount = 0;
optscore = score;

for ii = 200:-1:1
    
    [Q R] = qr(U);
    Q = Q*sign(R(1,1));
    U = Q(:,1:m);
    V = Q(:,(m+1):end);
    
    u_x_ce = U'*x_ce;
    h = zeros(1,dim);
    u_x_nu = U'*x_nu;
    u_x_de = U'*x_de;
    
    %Compute first derivative
    dPd1 = zeros(dim,m);
    Ktemp = kernel_Gaussian(u_x_nu,u_x_ce,sigmat);
    for ll = 1:nBasis
        temp = -(u_x_nu - u_x_ce(:,ll)*ones(1,N_nu)).*(Ktemp(:,ll)*ones(1,m))';
        temp1 = (x_nu - x_ce(:,ll)*ones(1,N_nu));
        
        h = temp*temp1';
        
        dPd1 = dPd1 + h'*alphah(ll);
    end
    
    dPd1 = dPd1'/N_nu/sigma;
    
    %Compute second derivative
    Ktemp = kernel_Gaussian(u_x_de,u_x_ce,sigmat);
    dPd2 = zeros(size(U));
    for ll = 1:nBasis
        temp11 = (u_x_de - u_x_ce(:,ll)*ones(1,size(u_x_de,2))).*(ones(size(U,2),1)*Ktemp(:,ll)');
        tempx1 = (x_de - x_ce(:,ll)*ones(1,size(x_de,2)));
        for jj = 1:size(U,2)
            T11 = (temp11(jj,:)'*ones(1,nBasis)).*Ktemp;
            dPd2(:,jj) = dPd2(:,jj) - tempx1*T11*alphah*alphah(ll)*2;
        end
    end
    
    dPd2 = dPd2'/N_de/sigma;
    
    % Total derivative
    dPd = dPd1 - 0.5*dPd2;
    
    % Subspace Rotation
    dM  = [zeros(m,m) dPd*V;-(dPd*V)' zeros(dim-m,dim-m)];
    
    M   = 0.005*ii*dM;
    eM  = expm(M);
    U = [eye(m) zeros(m,dim-m)]*eM*[U'; V'];
    U = U';
    
    % Compute median score
    temp = U'*x_nu;
    med = compmedDist(temp');
    
    [alphah,sigmat,PD,wh_x_de,score]=uLSIF(U'*x_de,U'*x_nu,U'*x_ce,med*[0.75 1.0 1.25]);
    sigma = sigmat^2;
    
    if PD > PD_opt
        U_opt = U;
        alphah_opt = alphah;
        PD_opt = PD;
        sigmatopt = sigmat;
        optscore = score;
        decPDcount = decPDcount - 1;
        decPDcount = max(0,decPDcount);
    else
        decPDcount = decPDcount + 1;
    end
    
    %Stopping criterion
    if decPDcount > 20
        break;
    end
end


