function  lhss_param = LHSS_train(x_de, x_nu, nBasis,max_cv_dim)
%
% Least-squares Hetero-distributional Subspace Search (LHSS)
%
% Estimating ratio of probability densities
%   \frac{ p_{nu}(x) }{ p_{de}(x) }
% from samples
%    { xde_i | xde_i\in R^{d} }_{i=1}^{n_{de}}
% drawn independently from p_{de}(x) and samples
%    { xnu_i | xnu_i\in R^{d} }_{i=1}^{n_{nu}}
% drawn independently from p_{nu}(x).
%
% Usage:
%       lhss_param = LHSS_train(x_de,x_nu,param)
%
% Input:
%    x_de:          d by n_de sample matrix corresponding to `denominator' (iid from density p_de)
%    x_nu:          d by n_nu sample matrix corresponding to `numerator'   (iid from density p_nu)
%    nBasis:        Number of basis functions in uLSIF
%
%
% Output:
%    lhss_param.
%    nBasis:        Number of basis functions in uLSIF
%    U:             Transformation matrix to Hetero-distributional subspace
%    sigma:         Gaussian width in uLSIF
%    alpha:         uLSIF parameters
%    m:             Dimentionality selected by CV
%    x_ce:          Basis vectors
%    score:         CV score
%
% (c) Makoto Yamada, Department of Compter Science, Tokyo Institute of Technology, Japan.
%     yamada@sg.cs.titech.ac.jp,
%     http://sugiyama-www.cs.titech.ac.jp/~sugi/software/uLSIF/

if nargin<2
    error('number of input arguments is not enough!!!')
end

[dim,n_de]=size(x_de);
[dim_nu,n_nu]=size(x_nu);
if dim~=dim_nu
    error('dimension of two samples are diferent!!!')
end

% set number of basis functions
if nargin< 3
    nBasis = min(100, n_nu);
end

if nargin< 4
    max_cv_dim = 5;
end


%Estimate the optimal reduced dimensionality based on CV
mindim = min(dim,max_cv_dim);
score = zeros(1,mindim);
score_min = 100;
for mm = 1:mindim
    [U_opt PD_opt sigmatopt alphah_opt x_ce score(mm)] = LHSS(x_de, x_nu,mm,nBasis);
    
    if score_min > score(mm)
        optm = mm;
        score_min = score(mm);
    end
end

%Compute LHSS based on the estimated dimension.
[U_opt PD_opt sigmatopt alphah_opt x_ce] = LHSS(x_de, x_nu, optm, nBasis);

% set LHSS parameters
lhss_param.nBasis = nBasis;
lhss_param.U = U_opt;
lhss_param.sigma = sigmatopt;
lhss_param.alpha = alphah_opt;
lhss_param.m     = optm;
lhss_param.x_ce  = x_ce;
lhss_param.score = score;

